import assess
import math
import numpy as np
import scipy

# Y(), dYdphi(), and dYdtheta() depend on scipy sph_harm_y
# test whether the conventions correspond to what I assumed:
# * phi is longitude [0, 2pi]
# * theta is co-latitude [0,pi]
# As that went wrong when changing from sph_harm -> sph_harm_y
# sph_harm_y uses now uses the same coordinate convention as assess
# (where sph_harm didn't)
# Further we use:
# * l for degree (n in scipy)
# * m for order  (same in scipy


def assumed_normalisation(l, m):
    # based on scipy docstring
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.sph_harm_y.html
    return np.sqrt((2*l+1)/(4*math.pi))*np.sqrt(math.factorial(l-m)/math.factorial(l+m))


def test_Y():
    for phi in np.linspace(0, 2*math.pi, 10):
        for theta in np.linspace(0, math.pi, 10):
            for l in range(5):
                for m in range(0, l):
                    N = assumed_normalisation(l, m)
                    # NOTE: lpmv takes order, degree (i.e. m, l)
                    P = scipy.special.lpmv(m, l, math.cos(theta))
                    expected = N * P * math.cos(m*phi)
                    Y = assess.Y(l, m, theta, phi)
                    np.testing.assert_allclose(Y, expected, atol=1e-15)



"""Assess is a python package that implements analytical solutions to the Stokes equations
in cylindrical and spherical domains.
"""
__all__ = ['CylindricalStokesSolutionSmoothFreeSlip', 'CylindricalStokesSolutionSmoothZeroSlip',  # noqa:F405
           'CylindricalStokesSolutionDeltaFreeSlip', 'CylindricalStokesSolutionDeltaZeroSlip',  # noqa:F405
           'SphericalStokesSolutionSmoothFreeSlip', 'SphericalStokesSolutionSmoothZeroSlip',  # noqa:F405
           'SphericalStokesSolutionDeltaFreeSlip', 'SphericalStokesSolutionDeltaZeroSlip',  # noqa:F405
           'Y', 'dYdphi', 'dYdtheta', 'to_spherical', 'from_spherical']  # noqa:F405
from .cylindrical import *  # noqa:F403,F401
from .spherical import *  # noqa:F403,F401



"""Analytical solutions in spherical shell domains."""
from __future__ import division
from assess.cylindrical import AnalyticalStokesSolution
from assess.smooth import coefficients_sphere_smooth_fs, coefficients_sphere_smooth_ns
from assess.delta import coefficients_sphere_delta_fs, coefficients_sphere_delta_ns
import scipy.special
import numpy
from math import sqrt, atan2, cos, sin, tan, acos, pi

__all__ = ['SphericalStokesSolutionSmoothFreeSlip', 'SphericalStokesSolutionSmoothZeroSlip',
           'SphericalStokesSolutionDeltaFreeSlip', 'SphericalStokesSolutionDeltaZeroSlip',
           'Y', 'dYdphi', 'dYdtheta', 'Y_cartesian', 'to_spherical', 'from_spherical']


# Legendre polynomials and their derivatives:

def Y(l, m, theta, phi):
    r"""Real-valued spherical harmonic function :math:`Y_{lm}(\theta, \varphi)`

    This is based on the following definition:

    .. math::

        Y_{lm}(\theta, \varphi) = \sqrt{\frac{(2l+1)}{4\pi}\frac{(l-m)!}{(l+m)!}}P_l^m(\cos(\theta))\cos(m\varphi)

    which is equal to the real part of
    `scipy.special.sph_harm_y <https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.sph_harm_y.html>`_.

    :param l: degree of the harmonic
    :param m: order of the harmonic
    :param theta: co-latitude in [0, pi]
    :param phi: longitude in [0, 2*pi]
    """
    # everywhere we take the real part of Y, corresponding to the cos(m phi) part of the solution
    return scipy.special.sph_harm_y(l, m, theta, phi).real


def dYdphi(l, m, theta, phi):
    r"""Colatitudinal derivative of spherical harmonic function :math:`Y_{lm}(\theta, \varphi)`

    :param l: degree of the harmonic
    :param m: order of the harmonic
    :param theta: co-latitude in [0, pi]
    :param phi: longitude in [0, 2*pi]
    """
    # except in theta derivatives (of odd order)
    return -m * scipy.special.sph_harm_y(l, m, theta, phi).imag


def dYdtheta(l, m, theta, phi):
    r"""Longitudinal derivative of spherical harmonic function :math:`Y_{lm}(\theta, \varphi)`


    :param l: degree of the harmonic
    :param m: order of the harmonic
    :param theta: co-latitude in [0, pi]
    :param phi: longitude in [0, 2*pi]
    """
    # this is from http://functions.wolfram.com/Polynomials/SphericalHarmonicY/20/01/01/
    # which can be derived from (x^2-1)d/dx P^ml = sqrt(1-x^2) P^(m+1)l -mx P^ml
    dydt = m/tan(theta) * Y(l, m, theta, phi)
    if m < l:
        # for m==l, Y^(m+1)_l=0
        # note we fiddle with phi to obtain the desired exp(im phi)
        # despite raising m to m+1
        dydt += sqrt((l-m)*(l+m+1)) * Y(l, m+1, theta, phi*m/(m+1))
    return dydt


def to_spherical(X):
    """Convert Cartesian 3D coordinates X to spherical r, theta, phi

    :param X: Cartesian 3D coordinates
    :returns r, theta, phi: radius, colatitude, longitude
    """
    r = sqrt(X[0]**2+X[1]**2+X[2]**2)
    theta = acos(X[2]/r)
    phi = atan2(X[1], X[0])
    return r, theta, phi


def from_spherical(r, theta, phi):
    """Convert spherical r, theta, phi to 3D Cartesian coordinates.

    :param r,: radius
    :param theta: co-latitude in [0, pi]
    :param phi: longitude in [0, 2*pi]
    :returns X: Cartesian 3D coordinates
    """
    return [r*cos(phi), r*cos(theta)*sin(phi), r*sin(theta)*sin(phi)]


def Y_cartesian(l, m, X):
    """Real-valued spherical harmonic function that takes Cartesian coordinates

    See :func:`Y`
    :param m: order of the harmonic
    :param l: degree of the harmonic
    :param X: Cartesian 3D coordinates
    """
    r, theta, phi = to_spherical(X)
    return Y(l, m, theta, phi)


class SphericalStokesSolution(AnalyticalStokesSolution):
    r"""Base class for solutions in spherical shell domains.

    This implements analytical solutions based on a poloidal function
    of the form.

    .. math::

        \mathcal{P}(r,\theta,\varphi) = \mathcal{P}_l(r)Y_{lm}(\theta, \varphi)

    so that the velocity solution takes the form:

    .. math::

        \mathbf{u} = \nabla\times\left(\mathbf{r}\times\nabla\mathcal{P}\right)

    The function :math:`\mathcal{P}_l` and its derivative should
    be implement in methods :meth:`Pl` and :meth:`dPldr` in the subclass.
    """
    def __init__(self, l, m, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        super(SphericalStokesSolution, self).__init__(Rp=Rp, Rm=Rm, nu=nu, g=g)
        if l < 0:
            raise ValueError("Spherical order l cannot be negative.")
        if abs(m) > l:
            raise ValueError("Absolute value of spherical degree m should be equal or smaller than order l")
        self.l = l
        self.m = m

    def Pl(self, r):
        """Abstract method to be implemented by subclass.

        :param r: radius
        """
        raise NotImplementedError("Should implement Pl() method")

    def dPldr(self, r):
        """Abstract method to be implemented by subclass.

        :param r: radius
        """
        raise NotImplementedError("Should implement dPldr() method")

    def u_theta(self, r, theta, phi):
        """Return colatitudinal (southward) component of velocity.

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        # u_theta = -1/r d/dtheta d/dr (r P)
        #         = -1/r dP/dtheta - d/dtheta d/dr P
        #         = -(1/r Pl + dPl/dr) dY/dtheta
        return -(self.Pl(r)/r + self.dPldr(r)) * dYdtheta(self.l, self.m, theta, phi)

    def u_phi(self, r, theta, phi):
        """Return longitudinal (eastward) component of velocity.

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        # u_phi = -1/(r sin(theta)) d/dphi d/dr (r P)
        #       = -1/(r sin(theta)) * (dP/dphi + r d/dphi d/dr P)
        #       = -1/sin(theta) * (Pl/r + dPl/dr) * dY/dphi
        return -(self.Pl(r)/r + self.dPldr(r)) / sin(theta) * dYdphi(self.l, self.m, theta, phi)

    def u_r(self, r, theta, phi):
        """Return radial component of velocity.

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        # u_r = 1/r \Lambda^2 P = -1/r l(l+1) P
        return -self.l*(self.l+1)*self.Pl(r)*Y(self.l, self.m, theta, phi)/r

    def tau_rr(self, r, theta, phi):
        """Return radial component of deviatoric stress.

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        # tau_rr = 2 nu d/dr (1/r \Lambda^2 P)
        #        = -2 nu l(l+1) [1/r dP/dr - 1/r^2 P]
        return -2*self.nu*self.l*(self.l+1)*(self.dPldr(r) - self.Pl(r)/r)*Y(self.l, self.m, theta, phi)/r

    def tau_rtheta(self, r, theta, phi):
        """Return colatitudinal component of shear stress.

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        # tau_rphi = \nu d/dtheta (1/r^2 \Lambda^2 P - d^2P/dr^2 + 2/r^2 P)
        l, m = self.l, self.m
        return self.nu * (-l*(l+1)*self.Pl(r)/r**2 - self.dPldr2(r) + 2*self.Pl(r)/r**2)*dYdtheta(l, m, theta, phi)

    def tau_rphi(self, r, theta, phi):
        """Return longitudinal component of shear stress.

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        # tau_rphi = \nu d/dphi (1/r^2 \Lambda^2 P - d^2P/dr^2 + 2/r^2 P)/sin(theta)
        l, m = self.l, self.m
        return self.nu * (-l*(l+1)*self.Pl(r)/r**2 - self.dPldr2(r) + 2*self.Pl(r)/r**2)*dYdphi(l, m, theta, phi)/sin(theta)

    def radial_stress(self, r, theta, phi):
        """Return radial component of stress.

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        return self.tau_rr(r, theta, phi) - self.p(r, theta, phi)

    def radial_stress_cartesian(self, X):
        """Return radial component of stress at Cartesian location.

        :param X: 3D Cartesian location"""
        return self.radial_stress(*to_spherical(X))

    def velocity_cartesian(self, X):
        """Return Cartesian velocity at Cartesian location.

        :param X: 3D Cartesian location"""
        r, theta, phi = to_spherical(X)
        if theta < 1e-7*pi or theta > pi*(1.-1e-7):
            # workaround pole problem by averaging in 4 points near the pole
            dx = 1e-6*r
            return tuple(numpy.mean([self.velocity_cartesian((x, y, X[2])) for x, y in [[dx, dx], [-dx, dx], [dx, -dx], [-dx, -dx]]], axis=0))
        ur = self.u_r(r, theta, phi)
        uth = self.u_theta(r, theta, phi)
        uph = self.u_phi(r, theta, phi)
        costh = cos(theta)
        req = sqrt(X[0]**2+X[1]**2)
        return (X[0]/r*ur + X[0]/req*costh*uth - X[1]/req*uph,
                X[1]/r*ur + X[1]/req*costh*uth + X[0]/req*uph,
                X[2]/r*ur - sin(theta)*uth)

    def pressure_cartesian(self, X):
        """Return pressure solution at Cartesian location.

        :param X: 3D Cartesian location"""
        return self.p(*to_spherical(X))


class SphericalStokesSolutionDelta(SphericalStokesSolution):
    r"""Base class for solutions in spherical shell domains with delta(r-r') forcing

    This implements the analytical solution in one half (above or below r')
    of the domain which is based on a poloidal function

    .. math ::

        \mathcal{P}(r,\theta,\varphi) = \mathcal{P}_l(r)Y_{lm}(\theta, \varphi)

    and velocity

    .. math ::

        \mathbf{u} = \nabla\times\left(\mathbf{r}\times\nabla\mathcal{P}\right)

    where for biharmonic solutions, :math:`\mathcal{P}_l(r)` is determinded by four
    coefficients, A, B, C, and D.
    """
    def __init__(self, ABCD, l, m, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """:param ABCD: list or tuple of 4 floats, coefficients for the 4 independent biharmonic
           solutions.
        :param l: degree of the harmonic
        :param m: order of the harmonic
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        super(SphericalStokesSolutionDelta, self).__init__(l, m, Rp=Rp, Rm=Rm, nu=nu, g=g)
        self.ABCD = ABCD
        A, B, C, D = self.ABCD
        self.G = -2*nu*(l+1)*(2*l+3)*C
        self.H = -2*nu*l*(2*l-1)*D

    def Pl(self, r):
        """Radial part of poloidal function

        :param r: radius"""
        A, B, C, D = self.ABCD
        l = self.l
        return A*r**l + B*r**(-l-1) + C*r**(l+2) + D*r**(-l+1)

    def dPldr(self, r):
        """Radial derivative of radial part of poloidal function

        :param r: radius"""
        A, B, C, D = self.ABCD
        l = self.l
        return l*A*r**(l-1) + (-l-1)*B*r**(-l-2) + (l+2)*C*r**(l+1) + (-l+1)*D*r**-l

    def dPldr2(self, r):
        """Second radial derivative of radial part of poloidal function

        :param r: radius"""
        A, B, C, D = self.ABCD
        l = self.l
        return l*(l-1)*A*r**(l-2) + (-l-1)*(-l-2)*B*r**(-l-3) + (l+2)*(l+1)*C*r**l + (-l+1)*(-l)*D*r**(-l-1)

    def p(self, r, theta, phi):
        """Pressure solution

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        l, m = self.l, self.m
        return (self.G*r**l + self.H*r**(-l-1))*Y(l, m, theta, phi)


class SphericalStokesSolutionSmooth(SphericalStokesSolution):
    r"""Base class for solutions in spherical shell domains with r^k forcing

    This implements the analytical solution in one half (above or below r')
    of the domain which is based on a poloidal function

    .. math ::

        \mathcal{P}(r,\theta,\varphi) = \mathcal{P}_l(r)Y_{lm}(\theta, \varphi)

    and velocity

    .. math ::

        \mathbf{u} = \nabla\times\left(\mathbf{r}\times\nabla\mathcal{P}\right)

    where the solution :math:`\mathcal{P}_l(r)` is determinded by four
    coefficients, A, B, C, and D of four independent biharmonic solutions,
    and one coefficient E associated with the inhomogenous part.
    """
    def __init__(self, ABCDE, k, l, m, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """:param ABCDE: list or tuple of 5 floats, coefficients for the 4 biharmonic
           and one inhomogenous solution.
        :param l: degree of the harmonic
        :param m: order of the harmonic
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        super(SphericalStokesSolutionSmooth, self).__init__(l, m, Rp=Rp, Rm=Rm, nu=nu, g=g)
        self.k = k
        self.ABCDE = ABCDE
        A, B, C, D, E = self.ABCDE
        self.G = -2*nu*(l+1)*(2*l+3)*C
        self.H = -2*nu*l*(2*l-1)*D
        self.K = -g*(k+2)/((k+1)*(k+2)-l*(l+1))/Rp**k

    def Pl(self, r):
        """Radial part of poloidal function

        :param r: radius"""
        A, B, C, D, E = self.ABCDE
        l, k = self.l, self.k
        return A*r**l + B*r**(-l-1) + C*r**(l+2) + D*r**(-l+1) + E*r**(k+3)

    def dPldr(self, r):
        """Radial derivative of radial part of poloidal function

        :param r: radius"""
        A, B, C, D, E = self.ABCDE
        l, k = self.l, self.k
        return l*A*r**(l-1) + (-l-1)*B*r**(-l-2) + (l+2)*C*r**(l+1) + (-l+1)*D*r**-l + (k+3)*E*r**(k+2)

    def dPldr2(self, r):
        """Second radial derivative of radial part of poloidal function

        :param r: radius"""
        A, B, C, D, E = self.ABCDE
        l, k = self.l, self.k
        return l*(l-1)*A*r**(l-2) + (-l-1)*(-l-2)*B*r**(-l-3) + (l+2)*(l+1)*C*r**l + (-l+1)*(-l)*D*r**(-l-1) + (k+3)*(k+2)*E*r**(k+1)

    def p(self, r, theta, phi):
        """Pressure solution

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        k, l, m = self.k, self.l, self.m
        return (self.G*r**l + self.H*r**(-l-1) + self.K*r**(k+1))*Y(l, m, theta, phi)

    def delta_rho(self, r, theta, phi):
        r"""Perturbation density :math:`\rho'` in forcing term: :math:`g\rho'\hat r`

        :param r: radius
        :param theta: co-latitude in [0, pi]
        :param phi: longitude in [0, 2*pi]
        """
        k, l, m = self.k, self.l, self.m
        return r**k * Y(l, m, theta, phi) / self.Rp**k

    def delta_rho_cartesian(self, X):
        r"""Perturbation density :math:`\rho'` in forcing term: :math:`g\rho'\hat r`

        :param X: 3D Cartesian coordinate"""
        return self.delta_rho(*to_spherical(X))


class SphericalStokesSolutionSmoothFreeSlip(SphericalStokesSolutionSmooth):
    """Analytical Solution in cylindrical domain with smooth r^k forcing and free-slip boundary conditions"""
    def __init__(self, l, m, k, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """:param l: degree of the harmonic
        :param m: order of the harmonic
        :param k: polynomial order of forcing
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        if (k+1)*(k+2) == l*(l+1) or (k+3)*(k+4) == l*(l+1):
            raise NotImplementedError("Smooth solution not implemented for k={}, l={}".format(k, l))
        ABCDE = coefficients_sphere_smooth_fs(Rp, Rm, k, l, g, nu)
        super(SphericalStokesSolutionSmoothFreeSlip, self).__init__(ABCDE, k, l, m, Rp=Rp, Rm=Rm, nu=nu, g=g)


class SphericalStokesSolutionSmoothZeroSlip(SphericalStokesSolutionSmooth):
    """Analytical Solution in cylindrical domain with smooth r^k forcing and zero-slip boundary conditions"""
    def __init__(self, l, m, k, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """:param l: degree of the harmonic
        :param m: order of the harmonic
        :param k: polynomial order of forcing
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        if (k+1)*(k+2) == l*(l+1) or (k+3)*(k+4) == l*(l+1):
            raise NotImplementedError("Smooth solution not implemented for k={}, l={}".format(k, l))
        ABCDE = coefficients_sphere_smooth_ns(Rp, Rm, k, l, g, nu)
        super(SphericalStokesSolutionSmoothZeroSlip, self).__init__(ABCDE, k, l, m, Rp=Rp, Rm=Rm, nu=nu, g=g)


class SphericalStokesSolutionDeltaFreeSlip(SphericalStokesSolutionDelta):
    """Analytical Solution in cylindrical domain with delta(r-r') forcing and free-slip boundary conditions"""
    def __init__(self, l, m, sign, Rp=2.22, Rm=1.22, rp=1.72, nu=1.0, g=1.0):
        r""":param l: degree of the harmonic
        :param m: order of the harmonic
        :param sign: +1 for upper half solution r'<r<Rp
                     -1 for lower half solution Rm<r<r'
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        ABCD = coefficients_sphere_delta_fs(Rp, Rm, rp, l, g, nu, sign)
        super(SphericalStokesSolutionDeltaFreeSlip, self).__init__(ABCD, l, m, Rp=Rp, Rm=Rm, nu=nu, g=g)


class SphericalStokesSolutionDeltaZeroSlip(SphericalStokesSolutionDelta):
    """Analytical Solution in cylindrical domain with delta(r-r') forcing and zero-slip boundary conditions"""
    def __init__(self, l, m, sign, Rp=2.22, Rm=1.22, rp=1.72, nu=1.0, g=1.0):
        r""":param l: degree of the harmonic
        :param m: order of the harmonic
        :param sign: +1 for upper half solution r'<r<Rp
                     -1 for lower half solution Rm<r<r'
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        ABCD = coefficients_sphere_delta_ns(Rp, Rm, rp, l, g, nu, sign)
        super(SphericalStokesSolutionDeltaZeroSlip, self).__init__(ABCD, l, m, Rp=Rp, Rm=Rm, nu=nu, g=g)



r"""This module contains functions that compute the coefficients in analytical
solution to the Stokes equations in cylindrical and spherical shell domains
with a smooth RHS forcing term, which in the 2D cylindrical shell domain takes
the form:

.. math::

  f = -g (r/Rp)^k cos(n\varphi) \hat{r}

where r is the radius (distance from origin), :math:`\varphi` is the angle with
the x-axis, Rp and Rm are the outer and inner radius of the shell domain, and
:math:`\hat{r}` is the outward pointing radial unit vector. The magnitude g,
degree l and wave number n can be chosen arbitrarily. Similarly in a 3D
spherical domain we consider the forcing term

.. math::

  f = -g (r/Rp)^k Y_{lm}(\theta,\varphi) \hat{r}

where :math:`\theta` and :math:`\varphi` are the co-latitude and longitude
respectively, and :math:`Y_{lm}` is Laplace's spherical harmonic function of
degree l and order m.

We consider the following cases:

  cylinder_smooth_fs: cylindrical, with free-slip boundary condition at r=Rm and r=Rp
  cylinder_smooth_ns: cylindrical, with zero-slip boundary conditions at r=Rm and r=Rp
  sphere_smooth_fs:   spherical, with free-slip boundary condition at r=Rm and r=Rp
  sphere_smooth_ns:    spherical, with zero-slip boundary condition at r=Rm and r=Rp

The functions below take the following parameters:

   :param Rp:   outer radius
   :param Rm:   inner radius
   :param k:    polynomial degree of forcing term
   :param n:    wave number (2D cylindrical only)
   :param l:    spherical degree (3D only)
   :param g:    scalar magnitude of forcing
   :param nu:   viscosity

and return the five coefficients A, B, C, D, and E used in the analytical solution.

This module has been automatically generated by extracting the solutions from the
latex source of the associated paper.
"""


def coefficients_cylinder_smooth_fs(Rp, Rm, k, n, g, nu):
    alpha = Rm/Rp
    A = -0.25*(alpha**2 - alpha**(k + n + 3))*Rp**(-n + 3)*g/((alpha + alpha**n)*(alpha**n - alpha)*(k + n + 1)*(k - n + 3)*nu)
    B = 0.25*Rp**(n + 3)*(alpha**(k + n + 3) - alpha**(2*n + 2))*g/((alpha**(n + 1) + 1)*(alpha**(n + 1) - 1)*(k + n + 3)*(k - n + 1)*nu)
    C = -0.25*Rp**(-n + 1)*(alpha**(k + n + 3) - 1)*g/((alpha**(n + 1) + 1)*(alpha**(n + 1) - 1)*(k + n + 3)*(k - n + 1)*nu)
    D = -0.25*Rp**(n + 1)*(alpha**(k + n + 3) - alpha**(2*n))*g/((alpha + alpha**n)*(alpha**n - alpha)*(k + n + 1)*(k - n + 3)*nu)
    E = g*n/(((k + 3)**2 - n**2)*((k + 1)**2 - n**2)*Rp**k*nu)
    return A, B, C, D, E


def coefficients_cylinder_smooth_ns(Rp, Rm, k, n, g, nu):
    alpha = Rm/Rp
    A = 0.5*((alpha**(k + n + 3) + alpha**(2*n))*(k + n + 1)*(n + 1) - (alpha**(k + n + 1) + alpha**(2*n + 2))*(k + n + 3)*n - (alpha**(k + 3*n + 3) + 1)*(k - n + 1))*Rp**(-n + 3)*g*n/(((alpha**(n + 1) - alpha**(n - 1))**2*n**2 - (alpha**(2*n) - 1)**2)*((k + 3)**2 - n**2)*((k + 1)**2 - n**2)*nu)
    B = -0.5*((alpha**(k + 3*n + 3) + alpha**(2*n))*(k - n + 1)*(n - 1) - (alpha**(k + 3*n + 1) + alpha**(2*n + 2))*(k - n + 3)*n + (alpha**(k + n + 3) + alpha**(4*n))*(k + n + 1))*Rp**(n + 3)*g*n/(((alpha**(n + 1) - alpha**(n - 1))**2*n**2 - (alpha**(2*n) - 1)**2)*((k + 3)**2 - n**2)*((k + 1)**2 - n**2)*nu)
    C = 0.5*((alpha**(k + n + 1) + alpha**(2*n))*(k + n + 3)*(n - 1) - (alpha**(k + n + 3) + alpha**(2*n - 2))*(k + n + 1)*n + (alpha**(k + 3*n + 1) + 1)*(k - n + 3))*Rp**(-n + 1)*g*n/(((alpha**(n + 1) - alpha**(n - 1))**2*n**2 - (alpha**(2*n) - 1)**2)*((k + 3)**2 - n**2)*((k + 1)**2 - n**2)*nu)
    D = -0.5*((alpha**(k + 3*n + 1) + alpha**(2*n))*(k - n + 3)*(n + 1) - (alpha**(k + 3*n + 3) + alpha**(2*n - 2))*(k - n + 1)*n - (alpha**(k + n + 1) + alpha**(4*n))*(k + n + 3))*Rp**(n + 1)*g*n/(((alpha**(n + 1) - alpha**(n - 1))**2*n**2 - (alpha**(2*n) - 1)**2)*((k + 3)**2 - n**2)*((k + 1)**2 - n**2)*nu)
    E = g*n/(((k + 3)**2 - n**2)*((k + 1)**2 - n**2)*Rp**k*nu)
    return A, B, C, D, E


def coefficients_sphere_smooth_fs(Rp, Rm, k, l, g, nu):
    alpha = Rm/Rp
    A = 0.5*Rp**(-l + 3)*(alpha**(k + 3) - alpha**(-l + 1))*g/((alpha**l - alpha**(-l + 1))*(k + l + 2)*(k - l + 3)*(2*l + 1)*nu)
    B = 0.5*Rp**(l + 4)*(alpha**(k + 4) - alpha**(l + 3))*g/((alpha**(l + 3) - 1/alpha**l)*(k + l + 4)*(k - l + 1)*(2*l + 1)*nu)
    C = -0.5*Rp**(-l + 1)*(alpha**(k + 4) - 1/alpha**l)*g/((alpha**(l + 3) - 1/alpha**l)*(k + l + 4)*(k - l + 1)*(2*l + 1)*nu)
    D = -0.5*Rp**(l + 2)*(alpha**(k + 3) - alpha**l)*g/((alpha**l - alpha**(-l + 1))*(k + l + 2)*(k - l + 3)*(2*l + 1)*nu)
    E = g/(Rp**k*(k + l + 4)*(k + l + 2)*(k - l + 3)*(k - l + 1)*nu)
    return A, B, C, D, E


def coefficients_sphere_smooth_ns(Rp, Rm, k, l, g, nu):
    alpha = Rm/Rp
    Gamma = ((alpha**(l + 1) + alpha**(l - 3))*(2*l + 1)**2 - 2*alpha**(l - 1)*(2*l + 3)*(2*l - 1) - 4*alpha**(3*l) - 4*alpha**(-l - 2))*(k + l + 4)*(k + l + 2)*(k - l + 3)*(k - l + 1)
    A = ((alpha**(k + 2) + alpha**(l - 1))*(k + l + 2)*(2*l + 3) - (alpha**k + alpha**(l + 1))*(k + l + 4)*(2*l + 1) - 2*(alpha**(k + 2*l + 3) + alpha**(-l - 2))*(k - l + 1))*Rp**(-l + 3)*g/(Gamma*nu)
    B = ((alpha**(k + 2*l + 1) + alpha**(l + 1))*(k - l + 3)*(2*l + 1) - (alpha**(k + 2*l + 3) + alpha**(l - 1))*(k - l + 1)*(2*l - 1) - 2*(alpha**(k + 2) + alpha**(3*l))*(k + l + 2))*Rp**(l + 4)*g/(Gamma*nu)
    C = -((alpha**(k + 2) + alpha**(l - 3))*(k + l + 2)*(2*l + 1) - (alpha**k + alpha**(l - 1))*(k + l + 4)*(2*l - 1) - 2*(alpha**(k + 2*l + 1) + alpha**(-l - 2))*(k - l + 3))*Rp**(-l + 1)*g/(Gamma*nu)
    D = -((alpha**(k + 2*l + 1) + alpha**(l - 1))*(k - l + 3)*(2*l + 3) - (alpha**(k + 2*l + 3) + alpha**(l - 3))*(k - l + 1)*(2*l + 1) - 2*(alpha**k + alpha**(3*l))*(k + l + 4))*Rp**(l + 2)*g/(Gamma*nu)
    E = g/(Rp**k*(k + l + 4)*(k + l + 2)*(k - l + 3)*(k - l + 1)*nu)
    return A, B, C, D, E



"""Analytical solutions in cylindrical domains."""
from __future__ import division
from assess.smooth import coefficients_cylinder_smooth_fs, coefficients_cylinder_smooth_ns
from assess.delta import coefficients_cylinder_delta_fs, coefficients_cylinder_delta_ns
from math import sqrt, atan2, cos, sin


__all__ = ['CylindricalStokesSolutionSmoothFreeSlip', 'CylindricalStokesSolutionSmoothZeroSlip',
           'CylindricalStokesSolutionDeltaFreeSlip', 'CylindricalStokesSolutionDeltaZeroSlip']


class AnalyticalStokesSolution(object):
    """Base class for solutions in sperical or cylindrical shell domains."""
    def __init__(self, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """
        :param Rp: outer radius
        :param Rm: inner radius
        :param g: magnitude of source term
        :param nu: viscosity
        """
        self.nu = nu
        self.g = g
        self.Rp = Rp
        self.Rm = Rm


class CylindricalStokesSolution(AnalyticalStokesSolution):
    r"""Base class for solutions in cylindrical shell domains.

    This implements analytical solutions based on a streamfunction
    of the form.

    .. math::

        \psi(r,\varphi) = \psi_r(r)\sin(n\varphi)

    where :math:`\psi_r` should be defined in a method :meth:`psi_r`
    and its derivative in :meth:`dpsi_rdr`.
    """
    def __init__(self, n, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """Initialize basic parameters of analytical solution in cylindrical domain."""
        super(CylindricalStokesSolution, self).__init__(Rp=Rp, Rm=Rm, nu=nu, g=g)
        if n <= 1:
            raise NotImplementedError("Cylindrical cases only implemented for n>1")
        self.n = n

    def psi_r(self, r):
        """Abstract method to be implemented by subclass.

        :param r: radius
        """
        raise NotImplementedError("Should implement psi_r() method")

    def dpsi_rdr(self, r):
        """Abstract method to be implemented by subclass.

        :param r: radius"""
        raise NotImplementedError("Should implement dpsi_rdr() method")

    def u_r(self, r, phi):
        """Return radial component of velocity.

        :param r: radius
        :param phi: angle with x-axis."""
        dpsi_dphi = self.n*cos(self.n*phi)*self.psi_r(r)
        return -dpsi_dphi/r

    def u_phi(self, r, phi):
        """Return tangential component of velocity.

        :param r: radius
        :param phi: angle with x-axis."""
        dpsi_dr = sin(self.n*phi)*self.dpsi_rdr(r)
        return dpsi_dr

    def tau_rr(self, r, phi):
        """Return radial component of deviatoric stress.

        :param r: radius
        :param phi: angle with x-axis."""
        dpsi_dphi = self.n*cos(self.n*phi)*self.psi_r(r)
        dpsi_drdphi = self.n*cos(self.n*phi)*self.dpsi_rdr(r)
        return 2*self.nu*(dpsi_dphi/r**2 - dpsi_drdphi/r)

    def tau_rphi(self, r, phi):
        r"""Return shear stress :math:`\tau_{r\varphi}`.

        :param r: radius
        :param phi: angle with x-axis."""
        dpsi_dphi2 = -self.n**2 * self.psi_r(r)
        return self.nu*(self.dpsi_rdr2(r) - self.dpsi_rdr(r)/r - dpsi_dphi2/r**2)*sin(self.n*phi)

    def radial_stress(self, r, phi):
        """Return radial component of stress.

        :param r: radius
        :param phi: angle with x-axis."""
        return self.tau_rr(r, phi) - self.p(r, phi)

    def radial_stress_cartesian(self, X):
        """Return radial component of stress at Cartesian location.

        :param X: 2D Cartesian location"""
        r = sqrt(X[0]**2+X[1]**2)
        phi = atan2(X[1], X[0])
        return self.radial_stress(r, phi)

    def velocity_cartesian(self, X):
        """Return Cartesian velocity at Cartesian location.

        :param X: 2D Cartesian location"""
        r = sqrt(X[0]**2+X[1]**2)
        phi = atan2(X[1], X[0])
        ur = self.u_r(r, phi)
        ut = self.u_phi(r, phi)
        return [ur*X[0]/r - ut*X[1]/r, ur*X[1]/r + ut*X[0]/r]

    def pressure_cartesian(self, X):
        """Return pressure solution at Cartesian location.

        :param X: 2D Cartesian location"""
        r = sqrt(X[0]**2+X[1]**2)
        phi = atan2(X[1], X[0])
        return self.p(r, phi)


class CylindricalStokesSolutionDelta(CylindricalStokesSolution):
    r"""Base class for solutions in cylindrical shell domains with delta(r-r') forcing

    This implements the analytical solution in one half (above or below r')
    of the domain which is based on a biharmonic streamfunction

    .. math::

        \psi(r,\varphi) = \psi_r(r)\sin(n\varphi)

    determinded by 4 coefficients, A, B, C, and D.
    """
    def __init__(self, ABCD, n, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """:param ABCD: list or tuple of 4 floats, coefficients for the 4 biharmonic
           solutions of the streamfunction
        :param n: wave number
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        super(CylindricalStokesSolutionDelta, self).__init__(n, Rp=Rp, Rm=Rm, nu=nu, g=g)
        self.ABCD = ABCD
        A, B, C, D = self.ABCD
        self.G = -4*self.nu*C*(n+1)
        self.H = -4*self.nu*D*(n-1)

    def psi_r(self, r):
        """Radial part of streamfunction

        :param r: radius"""
        A, B, C, D = self.ABCD
        n = self.n
        return A*r**n+B*r**(-n)+C*r**(n+2)+D*r**(-n+2)

    def dpsi_rdr(self, r):
        """Radial derivative of radial part of streamfunction

        :param r: radius"""
        A, B, C, D = self.ABCD
        n = self.n
        return A*n*r**(n-1) + B*-n*r**(-n-1) + C*(n+2)*r**(n+1) + D*(-n+2)*r**(-n+1)

    def dpsi_rdr2(self, r):
        """Second radial derivative of radial part of streamfunction

        :param r: radius"""
        A, B, C, D = self.ABCD
        n = self.n
        return A*n*(n-1)*r**(n-2) + B*-n*(-n-1)*r**(-n-2) + C*(n+2)*(n+1)*r**n + D*(-n+2)*(-n+1)*r**(-n)

    def p(self, r, phi):
        """Pressure solution

        :param r: radius
        :param phi: angle with x-axis"""
        n = self.n
        return (self.G*r**n + self.H*r**(-n))*cos(n*phi)


class CylindricalStokesSolutionSmooth(CylindricalStokesSolution):
    r"""Base class for solutions in cylindrical shell domains with r^k forcing

    The analytical solution is based on a streamfunction

    .. math:

        \psi(r,\varphi) = \psi_r(r)\sin(n\varphi)

    determinded by 4 coefficients, A, B, C, and D corresponding to 4 independent
    biharmonic (i.e. homogenous) solutions and a fifth coefficient E that
    corresponds to the inhomogeneous part.
    """
    def __init__(self, ABCDE, k, n, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """:param ABCDE: list or tuple of 5 floats, coefficients for the 4 biharmonic
           and one inhomogenous solutions of the streamfunction
        :param k: polynomial order of forcing
        :param n: wave number
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        super(CylindricalStokesSolutionSmooth, self).__init__(n, Rp=Rp, Rm=Rm, nu=nu, g=g)
        self.k = k
        self.ABCDE = ABCDE
        A, B, C, D, E = self.ABCDE
        self.G = -4*self.nu*C*(n+1)
        self.H = -4*self.nu*D*(n-1)
        self.F = -g*(k + 1)*Rp**(-k)/((k+1)**2-n**2)

    def psi_r(self, r):
        """Radial part of streamfunction

        :param r: radius"""
        A, B, C, D, E = self.ABCDE
        n, k = self.n, self.k
        return A*r**n+B*r**(-n)+C*r**(n+2)+D*r**(-n+2)+E*r**(k+3)

    def dpsi_rdr(self, r):
        """Radial derivative of radial part of streamfunction

        :param r: radius"""
        A, B, C, D, E = self.ABCDE
        n, k = self.n, self.k
        return A*n*r**(n-1) + B*-n*r**(-n-1) + C*(n+2)*r**(n+1) + D*(-n+2)*r**(-n+1) + E*(k+3)*r**(k+2)

    def dpsi_rdr2(self, r):
        """Second radial derivative of radial part of streamfunction

        :param r: radius"""
        A, B, C, D, E = self.ABCDE
        n, k = self.n, self.k
        return A*n*(n-1)*r**(n-2) + B*-n*(-n-1)*r**(-n-2) + C*(n+2)*(n+1)*r**n + D*(-n+2)*(-n+1)*r**(-n) + E*(k+3)*(k+2)*r**(k+1)

    def p(self, r, phi):
        """Pressure solution

        :param r: radius
        :param phi: angle with x-axis"""
        n, k = self.n, self.k
        return (self.G*r**n + self.H*r**(-n) + self.F*r**(k+1))*cos(n*phi)

    def delta_rho(self, r, phi):
        r"""Perturbation density :math:`\rho'` in forcing term: :math:`g\rho'\hat r`

        :param r: radius
        :param phi: angle with x-axis"""
        n, k = self.n, self.k
        return r**k * cos(n*phi) / self.Rp**k

    def delta_rho_cartesian(self, X):
        r"""Perturbation density :math:`\rho'` in forcing term: :math:`g\rho'\hat r`

        :param X: 2D Cartesian coordinate"""
        r = sqrt(X[0]**2+X[1]**2)
        phi = atan2(X[1], X[0])
        return self.delta_rho(r, phi)


class CylindricalStokesSolutionSmoothFreeSlip(CylindricalStokesSolutionSmooth):
    """Analytical Solution in cylindrical domain with smooth r^k forcing and free-slip boundary conditions"""
    def __init__(self, n, k, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """:param n: wave number
        :param k: polynomial order of forcing
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        if abs(k+3) == n or abs(k+1) == n:
            raise NotImplementedError("Smooth solution not implemented for k={}, n={}".format(k, n))
        ABCDE = coefficients_cylinder_smooth_fs(Rp, Rm, k, n, g, nu)
        super(CylindricalStokesSolutionSmoothFreeSlip, self).__init__(ABCDE, k, n, Rp=Rp, Rm=Rm, nu=nu, g=g)


class CylindricalStokesSolutionSmoothZeroSlip(CylindricalStokesSolutionSmooth):
    """Analytical Solution in cylindrical domain with smooth r^k forcing and zero-slip boundary conditions"""
    def __init__(self, n, k, Rp=2.22, Rm=1.22, nu=1.0, g=1.0):
        """:param n: wave number
        :param k: polynomial order of forcing
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        if abs(k+3) == n or abs(k+1) == n:
            raise NotImplementedError("Smooth solution not implemented for k={}, n={}".format(k, n))
        ABCDE = coefficients_cylinder_smooth_ns(Rp, Rm, k, n, g, nu)
        super(CylindricalStokesSolutionSmoothZeroSlip, self).__init__(ABCDE, k, n, Rp=Rp, Rm=Rm, nu=nu, g=g)


class CylindricalStokesSolutionDeltaFreeSlip(CylindricalStokesSolutionDelta):
    """Analytical Solution in cylindrical domain with delta(r-r') forcing and free-slip boundary conditions"""
    def __init__(self, n, sign, Rp=2.22, Rm=1.22, rp=1.72, nu=1.0, g=1.0):
        r""":param n: wave number
        :param sign: +1 for upper half solution r'<r<Rp
                     -1 for lower half solution Rm<r<r'
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        ABCD = coefficients_cylinder_delta_fs(Rp, Rm, rp, n, g, nu, sign)
        super(CylindricalStokesSolutionDeltaFreeSlip, self).__init__(ABCD, n, Rp=Rp, Rm=Rm, nu=nu, g=g)


class CylindricalStokesSolutionDeltaZeroSlip(CylindricalStokesSolutionDelta):
    """Analytical Solution in cylindrical domain with delta(r-r') forcing and zero-slip boundary conditions"""
    def __init__(self, n, sign, Rp=2.22, Rm=1.22, rp=1.72, nu=1.0, g=1.0):
        r""":param n: wave number
        :param sign: +1 for upper half solution r'<r<Rp
                     -1 for lower half solution Rm<r<r'
        :param Rp: outer radius
        :param Rm: inner radius
        :param nu: viscosity
        :param g: forcing strength"""
        ABCD = coefficients_cylinder_delta_ns(Rp, Rm, rp, n, g, nu, sign)
        super(CylindricalStokesSolutionDeltaZeroSlip, self).__init__(ABCD, n, Rp=Rp, Rm=Rm, nu=nu, g=g)



r"""This module contains functions that compute the coefficients in analytical
solution to the Stokes equations in cylindrical and spherical shell domains
with a delta function RHS forcing term, which in the 2D cylindrical shell
domain takes the form:

.. math::

  f = -g \delta(r-r') cos(n\varphi) \hat{r}

where r is the radius (distance from origin), :math:`\varphi` is the angle with
the x-axis, :math:`r'` is the radius of the density anomaly, and
:math:`\hat{r}` is the outward pointing radial unit vector. The magnitude g,
degree l and wave number n can be chosen arbitrarily. Similarly in a 3D
spherical domain we consider the forcing term

.. math::

  f = -g \delta(r-r') Y_{lm}(\theta,\varphi) \hat{r}

where :math:`\theta` and :math:`\varphi` are the co-latitude and longitude
respectively, and :math:`Y_{lm}` is Laplace's spherical harmonic function of
degree l and order m.

We consider the following cases:

  cylinder_delta_fs: cylindrical, with free-slip boundary condition at r=Rm and r=Rp
  cylinder_delta_ns: cylindrical, with zero-slip boundary conditions at r=Rm and r=Rp
  sphere_delta_fs:   spherical, with free-slip boundary condition at r=Rm and r=Rp
  sphere_smooh_ns:    spherical, with zero-slip boundary condition at r=Rm and r=Rp

The functions below take the following parameters:

   :param Rp:   outer radius
   :param Rm:   inner radius
   :param rp:   radius of density anomaly with Rm<rp<Rp
   :param n:    wave number (2D cylindrical only)
   :param l:    spherical degree (3D only)
   :param g:    scalar magnitude of forcing
   :param nu:   viscosity
   :param sign: +1 or -1 for the upper (r>rp) and lower (r<rp) half of the solution

and return four coefficients A, B, C, D for the linear combination of
biharmonic spherical functions that constitute the analytical solution.

This module has been automatically generated by extracting the solutions from
the latex source of the associated paper. If sign=+1 we use the following
substitutions

  alpha_pm = Rp/rp, alpha_mp = Rm/rp, pm=+1, mp=-1

i.e. we take the top case for the symbols \alpha_\pm, \alpha_\mp, \pm, and \mp
in the paper to obtain :math:`A_+, B_+, C_+,` and :math:`D_+` corresponding to
the upper half of the domain. If sign=-1 we have

  alpha_pm = Rm/rp, alpha_mp = Rp/rp, pm=-1, mp=+1

to obtain the :math:`A_-, B_-, C_-,` and :math:`D_-` coefficients of the lower
half of the solution.
"""


def coefficients_cylinder_delta_fs(Rp, Rm, rp, n, g, nu, sign):
    alpha_pm, alpha_mp = [Rp/rp, Rm/rp][::int(sign)]
    pm = sign

    A = -0.125*(alpha_mp**(2*n - 2) - 1)*g*pm*rp**(-n + 2)/((alpha_mp**(2*n - 2) - alpha_pm**(2*n - 2))*(n - 1)*nu)
    B = -0.125*(alpha_mp**(2*n + 2) - 1)*alpha_pm**(2*n + 2)*g*pm*rp**(n + 2)/((alpha_mp**(2*n + 2) - alpha_pm**(2*n + 2))*(n + 1)*nu)
    C = 0.125*(alpha_mp**(2*n + 2) - 1)*g*pm/((alpha_mp**(2*n + 2) - alpha_pm**(2*n + 2))*(n + 1)*nu*rp**n)
    D = 0.125*(alpha_mp**(2*n - 2) - 1)*alpha_pm**(2*n - 2)*g*pm*rp**n/((alpha_mp**(2*n - 2) - alpha_pm**(2*n - 2))*(n - 1)*nu)
    return A, B, C, D


def coefficients_cylinder_delta_ns(Rp, Rm, rp, n, g, nu, sign):
    alpha_p, alpha_m = [Rp/rp, Rm/rp]
    alpha_pm, alpha_mp = [Rp/rp, Rm/rp][::int(sign)]
    pm, mp = sign, -sign

    A = -0.125*(((alpha_m**2 - alpha_p**2)*n - (n + 1)*pm + 1/alpha_m**(2*n) - 1/alpha_p**(2*n))*(n - 1) + (alpha_m**2/alpha_p**(2*n) - alpha_p**2/alpha_m**(2*n))*n + (n**2*(alpha_m/alpha_p)**(2*mp) - (alpha_m/alpha_p)**(2*n*pm))*pm)*g*rp**(-n + 2)/((n**2*(alpha_m/alpha_p - alpha_p/alpha_m)**2 - ((alpha_m/alpha_p)**n - 1/(alpha_m/alpha_p)**n)**2)*(n - 1)*nu)
    B = -0.125*(((alpha_m**2 - alpha_p**2)*n - (n - 1)*pm - alpha_m**(2*n) + alpha_p**(2*n))*(n + 1) - (alpha_m**2*alpha_p**(2*n) - alpha_m**(2*n)*alpha_p**2)*n + (n**2*(alpha_m/alpha_p)**(2*mp) - (alpha_m/alpha_p)**(2*mp*n))*pm)*g*rp**(n + 2)/((n**2*(alpha_m/alpha_p - alpha_p/alpha_m)**2 - ((alpha_m/alpha_p)**n - 1/(alpha_m/alpha_p)**n)**2)*(n + 1)*nu)
    C = -0.125*((n**2*(alpha_m/alpha_p)**(2*pm) - (alpha_m/alpha_p)**(2*n*pm))*mp - (mp*(n - 1) + n*(1/alpha_m**2 - 1/alpha_p**2) - 1/alpha_m**(2*n) + 1/alpha_p**(2*n))*(n + 1) - n*(1/(alpha_m**(2*n)*alpha_p**2) - 1/(alpha_m**2*alpha_p**(2*n))))*g/((n**2*(alpha_m/alpha_p - alpha_p/alpha_m)**2 - ((alpha_m/alpha_p)**n - 1/(alpha_m/alpha_p)**n)**2)*(n + 1)*nu*rp**n)
    D = -0.125*((n**2*(alpha_m/alpha_p)**(2*pm) - (alpha_m/alpha_p)**(2*mp*n))*mp - (mp*(n + 1) + n*(1/alpha_m**2 - 1/alpha_p**2) + alpha_m**(2*n) - alpha_p**(2*n))*(n - 1) + n*(alpha_m**(2*n)/alpha_p**2 - alpha_p**(2*n)/alpha_m**2))*g*rp**n/((n**2*(alpha_m/alpha_p - alpha_p/alpha_m)**2 - ((alpha_m/alpha_p)**n - 1/(alpha_m/alpha_p)**n)**2)*(n - 1)*nu)
    return A, B, C, D


def coefficients_sphere_delta_fs(Rp, Rm, rp, l, g, nu, sign):
    alpha_pm, alpha_mp = [Rp/rp, Rm/rp][::int(sign)]
    pm = sign

    A = -0.5*(alpha_mp**(2*l - 1) - 1)*g*pm*rp**(-l + 2)/((alpha_mp**(2*l - 1) - alpha_pm**(2*l - 1))*(2*l + 1)*(2*l - 1)*nu)
    B = -0.5*(alpha_mp**(-2*l - 3) - 1)*g*pm*rp**(l + 3)/((alpha_mp**(-2*l - 3) - alpha_pm**(-2*l - 3))*(2*l + 3)*(2*l + 1)*nu)
    C = 0.5*(alpha_mp**(2*l + 3) - 1)*g*pm/((alpha_mp**(2*l + 3) - alpha_pm**(2*l + 3))*(2*l + 3)*(2*l + 1)*nu*rp**l)
    D = 0.5*(alpha_mp**(-2*l + 1) - 1)*g*pm*rp**(l + 1)/((alpha_mp**(-2*l + 1) - alpha_pm**(-2*l + 1))*(2*l + 1)*(2*l - 1)*nu)
    return A, B, C, D


def coefficients_sphere_delta_ns(Rp, Rm, rp, l, g, nu, sign):
    alpha_p, alpha_m = [Rp/rp, Rm/rp]
    alpha_pm, alpha_mp = [Rp/rp, Rm/rp][::int(sign)]
    pm, mp = sign, -sign

    A = -0.5*(alpha_m**2 - alpha_p**2 - (2*l + 1)*mp*(alpha_m/alpha_p)**(2*mp)/(2*l - 1) - (2*l + 3)*pm/(2*l + 1) + 2*(alpha_m**(-2*l - 1) - alpha_p**(-2*l - 1))/(2*l + 1) + 2*(alpha_m**2*alpha_p**(-2*l - 1) - alpha_m**(-2*l - 1)*alpha_p**2)/(2*l - 1) - 4*pm*(alpha_m/alpha_p)**((2*l + 1)*pm)/((2*l + 1)*(2*l - 1)))*g*rp**(-l + 2)/(((2*l + 1)**2*(alpha_m**2/alpha_p**2 + alpha_p**2/alpha_m**2) - 2*(2*l + 3)*(2*l - 1) - 4*(alpha_m/alpha_p)**(2*l + 1) - 4*(alpha_m/alpha_p)**(-2*l - 1))*nu)
    B = -0.5*(alpha_m**2 - alpha_p**2 - (2*l + 1)*mp*(alpha_m/alpha_p)**(2*mp)/(2*l + 3) - (2*l - 1)*pm/(2*l + 1) - 2*(alpha_m**2*alpha_p**(2*l + 1) - alpha_m**(2*l + 1)*alpha_p**2)/(2*l + 3) - 2*(alpha_m**(2*l + 1) - alpha_p**(2*l + 1))/(2*l + 1) - 4*pm*(alpha_m/alpha_p)**((2*l + 1)*mp)/((2*l + 3)*(2*l + 1)))*g*rp**(l + 3)/(((2*l + 1)**2*(alpha_m**2/alpha_p**2 + alpha_p**2/alpha_m**2) - 2*(2*l + 3)*(2*l - 1) - 4*(alpha_m/alpha_p)**(2*l + 1) - 4*(alpha_m/alpha_p)**(-2*l - 1))*nu)
    C = 0.5*((2*l + 1)*pm*(alpha_m/alpha_p)**(2*pm)/(2*l + 3) + (2*l - 1)*mp/(2*l + 1) - 2*(alpha_m**(-2*l - 1) - alpha_p**(-2*l - 1))/(2*l + 1) + 4*mp*(alpha_m/alpha_p)**((2*l + 1)*pm)/((2*l + 3)*(2*l + 1)) + 2*(alpha_m**(-2*l - 1)/alpha_p**2 - alpha_p**(-2*l - 1)/alpha_m**2)/(2*l + 3) + 1/alpha_m**2 - 1/alpha_p**2)*g/(((2*l + 1)**2*(alpha_m**2/alpha_p**2 + alpha_p**2/alpha_m**2) - 2*(2*l + 3)*(2*l - 1) - 4*(alpha_m/alpha_p)**(2*l + 1) - 4*(alpha_m/alpha_p)**(-2*l - 1))*nu*rp**l)
    D = 0.5*((2*l + 1)*pm*(alpha_m/alpha_p)**(2*pm)/(2*l - 1) + (2*l + 3)*mp/(2*l + 1) + 2*(alpha_m**(2*l + 1) - alpha_p**(2*l + 1))/(2*l + 1) + 4*mp*(alpha_m/alpha_p)**((2*l + 1)*mp)/((2*l + 1)*(2*l - 1)) - 2*(alpha_m**(2*l + 1)/alpha_p**2 - alpha_p**(2*l + 1)/alpha_m**2)/(2*l - 1) + 1/alpha_m**2 - 1/alpha_p**2)*g*rp**(l + 1)/(((2*l + 1)**2*(alpha_m**2/alpha_p**2 + alpha_p**2/alpha_m**2) - 2*(2*l + 3)*(2*l - 1) - 4*(alpha_m/alpha_p)**(2*l + 1) - 4*(alpha_m/alpha_p)**(-2*l - 1))*nu)
    return A, B, C, D



import pytest
import assess
from math import pi, cos, sin
from numpy.testing import assert_allclose

#######################
# the actual tests
######################


class _TestNoNormalFlow:
    """Test zero radial flow at top and bottom"""
    def test_no_normal_flow_top(self, case_top, angle):
        assert_allclose(case_top.u_r(case_top.Rp, *angle), 0, atol=1e-12)

    def test_no_normal_flow_bottom(self, case_bottom, angle):
        assert_allclose(case_bottom.u_r(case_bottom.Rm, *angle), 0, atol=1e-12)


class _TestFreeSlip(_TestNoNormalFlow):
    """Test zero shear stress in phi direction (and no normal from above)"""
    def test_free_slip_top(self, case_top, angle):
        assert_allclose(case_top.tau_rphi(case_top.Rp, *angle), 0, atol=1e-12)

    def test_free_slip_bottom(self, case_bottom, angle):
        assert_allclose(case_bottom.tau_rphi(case_bottom.Rm, *angle), 0, atol=1e-12)


class _TestSphericalFreeSlip(_TestFreeSlip):
    """Test zero shear stress in theta direction (and phi direction above, and no-normal flow)"""
    def test_free_slip_top_theta(self, case_top, angle):
        assert_allclose(case_top.tau_rtheta(case_top.Rp, *angle), 0, atol=1e-12)

    def test_free_slip_bottom_theta(self, case_bottom, angle):
        assert_allclose(case_bottom.tau_rtheta(case_bottom.Rm, *angle), 0, atol=1e-12)


class _TestZeroSlip(_TestNoNormalFlow):
    """Test zero velocity at top and bottom. Also seperately tests no-normal
    flow, so we test both u_r() and u_cartesian()"""
    def test_free_slip_top(self, case_top, top_X):
        assert_allclose(case_top.velocity_cartesian(top_X), 0, atol=1e-12)

    def test_free_slip_bottom(self, case_bottom, bottom_X):
        assert_allclose(case_bottom.velocity_cartesian(bottom_X), 0, atol=1e-12)


class _TestInterfaceConditions:
    """Test interface condition in r and phi directions."""
    def test_interface_conditions(self, case_top, case_bottom, angle, case_kwargs):
        rp = case_kwargs.get('rp', 1.72)
        assert_allclose(case_top.u_r(rp, *angle), case_bottom.u_r(rp, *angle))
        assert_allclose(case_top.u_phi(rp, *angle), case_bottom.u_phi(rp, *angle))
        assert_allclose(case_top.tau_rr(rp, *angle), case_bottom.tau_rr(rp, *angle))
        assert_allclose(case_top.tau_rphi(rp, *angle), case_bottom.tau_rphi(rp, *angle))


class _TestCylindricalInterfaceConditions(_TestInterfaceConditions):
    """Test interface condition in r and phi directions, including normal stress jump."""
    def test_interface_normal_stress(self, case_top, case_bottom, angle, case_kwargs):
        rp = case_kwargs.get('rp', 1.72)
        n, g = case_top.n, case_top.g
        phi = angle[0]
        assert_allclose(case_top.radial_stress(rp, *angle) - case_bottom.radial_stress(rp, *angle), g*cos(n*phi))


class _TestSphericalInterfaceConditions(_TestInterfaceConditions):
    """Test interface condition in r, phi and theta directions, including normal stress jump."""
    def test_interface_conditions_theta(self, case_top, case_bottom, angle, case_kwargs):
        rp = case_kwargs.get('rp', 1.72)
        assert_allclose(case_top.u_theta(rp, *angle), case_bottom.u_theta(rp, *angle))
        assert_allclose(case_top.tau_rr(rp, *angle), case_bottom.tau_rr(rp, *angle))
        assert_allclose(case_top.tau_rtheta(rp, *angle), case_bottom.tau_rtheta(rp, *angle))

    def test_interface_normal_stress(self, case_top, case_bottom, angle, case_kwargs):
        rp = case_kwargs.get('rp', 1.72)
        l, m, g = case_top.l, case_top.m, case_top.g
        assert_allclose(case_top.radial_stress(rp, *angle) - case_bottom.radial_stress(rp, *angle), g*assess.Y(l, m, *angle))


########################################################
# the fixtures that form the parameters for the tests
########################################################


class CylindricalFixtures:
    @pytest.fixture(params=[2, 3, 4, 5, 8])
    def n(self, request):
        return request.param

    @pytest.fixture(params=[0, pi/2, -pi/3])
    def angle(self, request):
        return [request.param]

    @pytest.fixture
    def top_X(self, case_top, angle):
        return [case_top.Rp * cos(angle[0]), case_top.Rp * sin(angle[0])]

    @pytest.fixture
    def bottom_X(self, case_bottom, angle):
        return [case_bottom.Rm * cos(angle[0]), case_bottom.Rm * sin(angle[0])]


class SphericalFixtures:
    @pytest.fixture(params=[2, 3, 4, 5, 8])
    def l(self, request):  # NOQA: E743
        return request.param

    @pytest.fixture(params=[-1, 0, 1])
    def m(self, request, l):
        return request.param*l

    @pytest.fixture(params=[[0.01, 0], [pi/2, 0], [pi-0.01, 0], [pi/2, 1.0], [1.5, 4.0]])
    def angle(self, request):
        return request.param

    @pytest.fixture
    def top_X(self, case_top, angle):
        return assess.from_spherical(case_top.Rp, *angle)

    @pytest.fixture
    def bottom_X(self, case_bottom, angle):
        return assess.from_spherical(case_bottom.Rm, *angle)


class SmoothFixtures:
    @pytest.fixture()
    def case_top(self, case):
        return case

    case_bottom = case_top

    @pytest.fixture(params=[
        {},
        {'Rm': 1.0, 'Rp': 2.0, 'nu': 3.0, 'g': 4.0},
        {'Rm': 0.9, 'Rp': 1.9, 'g': 9.81}])
    def case_kwargs(self, request):
        return request.param


class DeltaFixtures:
    @pytest.fixture(params=[
        {},
        {'rp': 2.},
        {'Rm': 1.0, 'rp': 1.2, 'Rp': 2.0, 'nu': 3.0, 'g': 4.0},
        {'Rm': 0.9, 'Rp': 1.9, 'g': 9.81}])
    def case_kwargs(self, request):
        return request.param


class SmoothCylindricalFixtures(CylindricalFixtures, SmoothFixtures):
    @pytest.fixture(params=[-7, -1, 0, 1, 2, 9])
    def k(self, request, n):
        k = request.param
        if (k+1)**2 == n**2 or (k+3)**2 == n**2:
            # NOTE: return None for not implemented combination
            return
        else:
            return k


class SmoothSphericalFixtures(SphericalFixtures, SmoothFixtures):
    @pytest.fixture(params=[-7, -1, 0, 1, 2, 9])
    def k(self, request, l):
        k = request.param
        if (k+1)*(k+2) == l*(l+1) or (k+3)*(k+4) == l*(l+1):
            # NOTE: return None for not implemented combination
            return
        else:
            return k


########################################################
# the 8 different cases
########################################################


class TestCylindricalStokesSolutionSmoothFreeSlip(SmoothCylindricalFixtures, _TestFreeSlip):
    @pytest.fixture()
    def case(self, n, k, case_kwargs):
        if k is None:
            pytest.xfail()
        return assess.CylindricalStokesSolutionSmoothFreeSlip(n, k, **case_kwargs)


class TestCylindricalStokesSolutionSmoothZeroSlip(SmoothCylindricalFixtures, _TestZeroSlip):
    @pytest.fixture()
    def case(self, n, k, case_kwargs):
        if k is None:
            pytest.xfail()
        return assess.CylindricalStokesSolutionSmoothZeroSlip(n, k, **case_kwargs)


class TestSphericalStokesSolutionSmoothFreeSlip(SmoothSphericalFixtures, _TestSphericalFreeSlip):
    @pytest.fixture()
    def case(self, l, m, k, case_kwargs):
        if k is None:
            pytest.xfail()
        return assess.SphericalStokesSolutionSmoothFreeSlip(l, m, k, **case_kwargs)


class TestSphericalStokesSolutionSmoothZeroSlip(SmoothSphericalFixtures, _TestZeroSlip):
    @pytest.fixture()
    def case(self, l, m, k, case_kwargs):
        if k is None:
            pytest.xfail()
        return assess.SphericalStokesSolutionSmoothZeroSlip(l, m, k, **case_kwargs)


class TestCylindricalStokesSolutionDeltaFreeSlip(CylindricalFixtures, DeltaFixtures, _TestFreeSlip, _TestCylindricalInterfaceConditions):
    @pytest.fixture()
    def case_top(self, n, case_kwargs):
        return assess.CylindricalStokesSolutionDeltaFreeSlip(n, +1, **case_kwargs)

    @pytest.fixture()
    def case_bottom(self, n, case_kwargs):
        return assess.CylindricalStokesSolutionDeltaFreeSlip(n, -1, **case_kwargs)


class TestCylindricalStokesSolutionDeltaZeroSlip(CylindricalFixtures, DeltaFixtures, _TestZeroSlip, _TestCylindricalInterfaceConditions):
    @pytest.fixture()
    def case_top(self, n, case_kwargs):
        return assess.CylindricalStokesSolutionDeltaZeroSlip(n, +1, **case_kwargs)

    @pytest.fixture()
    def case_bottom(self, n, case_kwargs):
        return assess.CylindricalStokesSolutionDeltaZeroSlip(n, -1, **case_kwargs)


class TestSphericalStokesSolutionDeltaFreeSlip(SphericalFixtures, DeltaFixtures, _TestSphericalFreeSlip, _TestSphericalInterfaceConditions):
    @pytest.fixture()
    def case_top(self, l, m, case_kwargs):
        return assess.SphericalStokesSolutionDeltaFreeSlip(l, m, +1, **case_kwargs)

    @pytest.fixture()
    def case_bottom(self, l, m, case_kwargs):
        return assess.SphericalStokesSolutionDeltaFreeSlip(l, m, -1, **case_kwargs)


class TestSphericalStokesSolutionDeltaZeroSlip(SphericalFixtures, DeltaFixtures, _TestZeroSlip, _TestSphericalInterfaceConditions):
    @pytest.fixture()
    def case_top(self, l, m, case_kwargs):
        return assess.SphericalStokesSolutionDeltaZeroSlip(l, m, +1, **case_kwargs)

    @pytest.fixture()
    def case_bottom(self, l, m, case_kwargs):
        return assess.SphericalStokesSolutionDeltaZeroSlip(l, m, -1, **case_kwargs)
